#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ggi/ggi.h>

typedef struct
{
  char signature;
  char version;
  char encoding;
  char bytes_per_pixel;
  unsigned short int xmin;
  unsigned short int ymin;
  unsigned short int xmax;
  unsigned short int ymax;
  unsigned short int vres;
  unsigned short int hres;
  char palette[48];
  char reserved;
  char color_layers;
  unsigned short int bytes_per_line;
  unsigned short int palette_type;
  char unused[58];
}PCX_Header;

unsigned char bytes_per_pixel;

void readpcximage(FILE * file,void * target,int size)
{
  unsigned char buf;
  unsigned int counter;
  int i=0;
  while(i<=size)  /* Image not entirely read? */
    {
      /* Get one byte */
      fread(&buf,1,1,file);
      /* Check the 2 most significant bits */
      if ((buf&192)==192)
	{
	  /* We have 11xxxxxx */
	  counter=(buf&63);    /* Number of times to repeat next byte */
	  fread(&buf,1,1,file);     /* Get next byte */
	  for(;counter>0;counter--) /* and copy it counter times */
	    {
	      ((char*)target)[i]=buf;
	      i++;   /* increase the number of bytes written */
	    }
	}
      else
	{
	  /* Just copy the byte */
	  ((char*)target)[i]=buf;
	  i++;   /* Increase the number of bytes written */
	}
    }
}

void *readpcx(FILE *file, char *palette,unsigned short int *length, 
unsigned short int *height)
     /* Returns NULL if failed, otherwise a pointer to the loaded image */
{
  PCX_Header header;
  void *target;
  fseek(file,0,SEEK_SET);
  fread(&header,sizeof(PCX_Header),1,file);   /* read the header */
  /* Check if this file is in pcx format */
  if((header.signature!=0x0a)||(header.version!=5))
    return(NULL);
  else
    {/* it is! */
      /* Return height and length */
      *length=header.xmax+1-header.xmin;
      *height=header.ymax+1-header.ymin;
      /* Allocate the sprite buffer */
      target=(void *)malloc((*length)*(*height));
      /* Read the image */
      readpcximage(file,target,(*length)*(*height));
      fseek(file,-768,SEEK_END);
      /* Get the palette */
      fread(palette,1,768,file);
      /* PCX succesfully read! */
      return(target);
    }
}

void init_ggi(ggi_visual_t * vis)
{
  ggi_mode sgmode;
  if (ggiInit()!=0) 
    {
      printf("Error initialising GGI!\n");
      exit(1);
    }
  if ((*vis=ggiOpen(NULL))==NULL) 
    {
      printf("Error opening visual!\n"); 
      ggiExit(); 
      exit(1);
    }
  /* Trying 8 bpp mode */
  if (!ggiCheckGraphMode(*vis,320,200,320,200,GT_8BIT,&sgmode))
    {
      if (ggiSetGraphMode(*vis,320,200,320,200,GT_8BIT))
	{
	  printf("Error initialising 8 bpp mode!\n"); 
	  exit(1);
	}
      /* Sucess */
      printf("Using 8bpp mode\n");
      bytes_per_pixel=1;
    }
  /* Can't use 8 bpp, trying 16... */
  else if (!ggiCheckGraphMode(*vis,320,200,320,200,GT_16BIT,&sgmode))
    {
      if (ggiSetGraphMode(*vis,320,200,320,200,GT_16BIT)) 
	{
	  printf("Error initialising 16 bpp mode!\n"); 
	  exit(1);
	}
      printf("Using 16bpp mode\n");
      bytes_per_pixel=2;
    }
  /* Neither 8 nor 16 work... Perhaps 24? */
  else if (!ggiCheckGraphMode(*vis,320,200,320,200,GT_24BIT,&sgmode))
    {
      if (ggiSetGraphMode(*vis,320,200,320,200,GT_24BIT)) 
	{
	  printf("Error initialising 24 bpp mode!!\n"); 
	  exit(1);
	}
      printf("Using 24bpp mode\n");
      bytes_per_pixel=3;
    }
  /* Here the user may have a 32bpp depth... But we don't handle this */
  else 
    {
      printf("No valid mode found!\n");
      exit(1);
    }
}

void exit_ggi(ggi_visual_t * vis)
{
  ggiClose(*vis);
  ggiExit();
}

void adapttodepth(ggi_visual_t vis, void * target, void * source, 
int lenght, char * palette, ggi_color * ggipalette)
/* Return in target a ready-to-display image from the 8bit source. 
   The corresponding palette must be set and target must be allocated */
{
  int i;
  switch(bytes_per_pixel)
    {
    case 1:
      /* 256 color mode, no adaptation needed */
      memcpy(target,source,lenght);
      break;
    case 2:
      /* 16bpp mode, we can use ggiMapColor */
      for(i=0;i<lenght;i++)
	((unsigned short int*)target)[i]=
          ggiMapColor(vis,&(ggipalette[((unsigned char*)source)[i]]));
      break;
    case 3:
      /* 24bpp mode, read the values directly from the palette */
      for(i=0;i<lenght;i++)
	{
	  ((unsigned char*)target)[(i*3)]=
            palette[(((unsigned char*)source)[i]*3)+2];
	  ((unsigned char*)target)[(i*3)+1]=
            palette[(((unsigned char*)source)[i]*3)+1];
	  ((unsigned char*)target)[(i*3)+2]=
            palette[(((unsigned char*)source)[i]*3)];
	}
      break;
    }
}

int main(int argc, char ** argv)
{
  ggi_visual_t vis;
  FILE * file;
  void * image;  /* 8 bit image as read from the pcx */
  void * trueimage;  /* Adapted to depth image */
  int i;
  unsigned short int lenght, height;
  unsigned char palette[768];
  ggi_color ggi_palette[768];
  if (argc!=2) 
    {
      printf("Usage: svgadisplaypcx filename.pcx.\n"); 
      return(1);
    }
  if ((file=fopen(argv[1],"r"))==NULL) 
    {
      printf("Can't open file!\n");
      return(1);
    }
  if ((image=readpcx(file,palette,&lenght,&height))==NULL)
    {
      printf("Error loading file!\n"); 
      return(1);
    }
  fclose(file);
  printf("Image is %dx%d sized.\n",lenght,height);
  if((lenght>320)||(height>200)) 
    {
      printf("Image is too big!\n");
      return(1);
    }
  for (i=0;i<256;i++)
    {
      ggi_palette[i].r=palette[i*3]<<8;
      ggi_palette[i].g=palette[(i*3)+1]<<8;
      ggi_palette[i].b=palette[(i*3)+2]<<8;
    }
  init_ggi(&vis);
  ggiSetPalette(vis,0,256,ggi_palette);
  trueimage=(void*)calloc(lenght*height,bytes_per_pixel);
  adapttodepth(vis, trueimage, image, lenght*height,palette,ggi_palette);
  ggiSetGCForeground(vis,0);
  ggiFillscreen(vis);
  ggiPutBox(vis,0,0,lenght,height,trueimage);
  ggiFlush(vis);
  ggiGetc(vis);
  /* Don't forget to free the allocated images! */
  free(trueimage);
  free(image);
  exit_ggi(&vis);
  return(0);
}
